Imports Microsoft.VisualBasic
Imports System.Windows
Imports System.Windows.Controls
Imports System.Windows.Data
Imports System.Windows.Documents
Imports System

Namespace FlowDocumentApp.FlowDocuments
	Friend NotInheritable Class Helpers
		''' <summary>
		''' If you use a bindable flow document element more than once, you may encounter a "Collection was modified" exception.
		''' The error occurs when the binding is updated because of a change to an inherited dependency property. The most common scenario 
		''' is when the inherited DataContext changes. It appears that an inherited properly like DataContext is propagated to its descendants. 
		''' When the enumeration of descendants gets to a BindableXXX, the dependency properties of that element change according to the new 
		''' DataContext, which change the (non-dependency) properties. However, for some reason, changing the flow content invalidates the 
		''' enumeration and raises an exception. 
		''' To work around this, one can either DataContext="{Binding DataContext, RelativeSource={RelativeSource AncestorType=FrameworkElement}}" 
		''' in code. This is clumsy, so every derived type calls this function instead (which performs the same thing).
		''' See http://code.logos.com/blog/2008/01/data_binding_in_a_flowdocument.html
		''' </summary>
		''' <param name="element"></param>
		Private Sub New()
		End Sub
		Public Shared Sub FixupDataContext(ByVal element As FrameworkContentElement)
			Dim b As New Binding(FrameworkContentElement.DataContextProperty.Name)
			' another approach (if this one has problems) is to bind to an ancestor by ElementName
			b.RelativeSource = New RelativeSource(RelativeSourceMode.FindAncestor, GetType(FrameworkElement), 1)
			element.SetBinding(FrameworkContentElement.DataContextProperty, b)
		End Sub


		Private Shared Function InternalUnFixupDataContext(ByVal dp As DependencyObject) As Boolean
			' only consider those elements for which we've called FixupDataContext(): they all belong to this namespace
			If TypeOf dp Is FrameworkContentElement AndAlso dp.GetType().Namespace = GetType(Helpers).Namespace Then
				Dim binding As Binding = BindingOperations.GetBinding(dp, FrameworkContentElement.DataContextProperty)
				If binding IsNot Nothing AndAlso binding.Path IsNot Nothing AndAlso binding.Path.Path = FrameworkContentElement.DataContextProperty.Name AndAlso binding.RelativeSource IsNot Nothing AndAlso binding.RelativeSource.Mode = RelativeSourceMode.FindAncestor AndAlso binding.RelativeSource.AncestorType Is GetType(FrameworkElement) AndAlso binding.RelativeSource.AncestorLevel = 1 Then
					BindingOperations.ClearBinding(dp, FrameworkContentElement.DataContextProperty)
					Return True
				End If
			End If
			' as soon as we have disconnected a binding, return. Don't continue the enumeration, since the collection may have changed
			For Each child As Object In LogicalTreeHelper.GetChildren(dp)
				If TypeOf child Is DependencyObject Then
					If InternalUnFixupDataContext(CType(child, DependencyObject)) Then
						Return True
					End If
				End If
			Next child
			Return False
		End Function


		Public Shared Sub UnFixupDataContext(ByVal dp As DependencyObject)
			Do While InternalUnFixupDataContext(dp)

			Loop
		End Sub


		''' <summary>
		''' Convert "data" to a flow document block object. If data is already a block, the return value is data recast.
		''' </summary>
		''' <param name="dataContext">only used when bindable content needs to be created</param>
		''' <param name="data"></param>
		''' <returns></returns>
		Public Shared Function ConvertToBlock(ByVal dataContext As Object, ByVal data As Object) As Block
			If TypeOf data Is Block Then
				Return CType(data, Block)
			ElseIf TypeOf data Is Inline Then
				Return New Paragraph(CType(data, Inline))
			ElseIf TypeOf data Is BindingBase Then
				Dim run As New BindableRun()
				If TypeOf dataContext Is BindingBase Then
					run.SetBinding(BindableRun.DataContextProperty, CType(dataContext, BindingBase))
				Else
					run.DataContext = dataContext
				End If
				run.SetBinding(BindableRun.BoundTextProperty, CType(data, BindingBase))
				Return New Paragraph(run)
			Else
				Dim run As New Run()
				run.Text = If((data Is Nothing), String.Empty, data.ToString())
				Return New Paragraph(run)
			End If
		End Function


		Public Shared Function LoadDataTemplate(ByVal dataTemplate As DataTemplate) As FrameworkContentElement
			Dim content As Object = dataTemplate.LoadContent()
			If TypeOf content Is Fragment Then
				Return CType((CType(content, Fragment)).Content, FrameworkContentElement)
			ElseIf TypeOf content Is TextBlock Then
				Dim inlines As InlineCollection = (CType(content, TextBlock)).Inlines
				If inlines.Count = 1 Then
					Return inlines.FirstInline
				Else
					Dim paragraph As New Paragraph()
					' we can't use an enumerator, since adding an inline removes it from its collection
					Do While inlines.FirstInline IsNot Nothing
						paragraph.Inlines.Add(inlines.FirstInline)
					Loop
					Return paragraph
				End If
			Else
				Throw New Exception("Data template needs to contain a <Fragment> or <TextBlock>")
			End If
		End Function
	End Class
End Namespace
